"""Functions for visualization purposes in Data Science tasks.
"""

import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

def plot_heatmap_between_columns(
        df: pd.DataFrame,
        size: tuple = (7, 7),
        savefig=False,
        path_and_file_name='heatmap') -> plt.axes:
    """Plot heatmap between values of multiple columns in a Pandas dataframe.
    Requires that there are no NaN values in the dataframe.

    Arguments:
        df {pd.DataFrame} -- Pandas dataframe
        size {tuple} -- Size of output figure

    Keyword Arguments:
        savefig {bool} -- True to save figure (default: {False})
        path_and_file_name {str} -- Filepath (default: {'heatmap'})

    Returns:
        plt.axes -- Matplotlib axes object
    """
    plt.figure(figsize=size)

    axes = sns.heatmap(df, vmin=-1, cmap='coolwarm', annot=True)

    if savefig:
        plt.savefig(path_and_file_name)

    return axes



from typing import List

import missingno as msno
import pandas as pd


from aeslib.data_science import interval_analysis as ia

def plot_correlation_between_missing_data(df: pd.DataFrame, group_by=None):
    """Get a Seaborn heatmap of column correlations.
    
    Arguments:
        df {pd.DataFrame} -- Pandas dataframe.
        group_by {str} -- Specify the name of a column in df to groupby if desired.
        Otherwise leave it as none. (default: {None})
    """
    if group_by is not None:
        grouped = df.groupby(group_by)
        for _, group in grouped:
            msno.heatmap(group)
    else:
        msno.heatmap(df)

def plot_missing_intervals(df: pd.DataFrame, group_by=None):
    """Plot missing intervals using functionality from the missingno package.
    
    Arguments:
        df {pd.DataFrame} -- Dataframe to plot missing data for.
    
    Keyword Arguments:
        group_by {str} -- Specify the name of a column in df to groupby if desired.
        Otherwise leave it as none. (default: {None})
    """
    if group_by is not None:
        grouped = df.groupby(group_by)
        for _, group in grouped:
            msno.matrix(group)
    else:
        msno.matrix(df)
    
def find_missing_intervals_for_dataframe(df: pd.DataFrame, cols: List = None) -> pd.DataFrame:
    """Returns a dataframe with all intervals where one or more columns contain a NaN value.
    The dataframe holds the start and end of the interval and it's duration. To examine a
    selection of columns pass in a list of columns in cols.
    
    Arguments:
        df {pd.DataFrame} -- A pandas dataframe with a single-level datetime
        index. Keep in mind that pd.to_datetime() must be formatted correctly.
        cols {List} -- A list of columns to examine. Default is None meaning all
        columns will be evaluated.
    
    Returns:
        pd.DataFrame -- Dataframe with intervals.
    """
    return ia.IntervalAnalysis.get_empty_intervals(df=df, cols=cols)



import datetime as dt
import json
import re
import requests
import pandas as pd
import numpy as np
import urllib3

from requests.auth import HTTPBasicAuth

from aeslib.data_engineering import azure_tools as at

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

def get_datetime_as_string(timestamp: dt.datetime.timestamp,
                           string_format="%Y_%m_%d_%H_%M_%S") -> str:
    time_stamp_as_datetime = dt.datetime.fromtimestamp(timestamp / 1000)
    return time_stamp_as_datetime.strftime(string_format)

def reduce_dataframe_size(df) -> pd.DataFrame:
    df.loc[:, df.select_dtypes(include=np.float64).columns] \
        = df.select_dtypes(include=np.float64).astype(np.float32)
    return df

def get_bazefield_authentication_from_key_vault(key_vault_name: str) -> HTTPBasicAuth:
    api_key = at.get_secret_from_key_vault(key_vault_name, secret_name='bazefield-api-key')
    return HTTPBasicAuth(api_key, '')

def get_bazefield_headers() -> dict:
    headers = {'Content-Type': 'application/json;charset=UTF-8',
               'Accept': 'application/json, text/plain, */*'}
    return headers

def get_full_tag_list_from_bazefield(key_vault_name: str):
    auth = get_bazefield_authentication_from_key_vault(key_vault_name)
    tag_list_url = at.get_secret_from_key_vault(key_vault_name, secret_name='bazefield-tagid-url')
    tag_list = requests.get(url=tag_list_url, verify=False, auth=auth)
    return json.loads(tag_list.text)

def match_taglist_to_regular_expression(tag_list: list, regular_expression: str) -> list:
    tag_ids = [t["tagId"] for t in tag_list
               if re.match(regular_expression, t["tagName"])]
    return tag_ids

def get_transformed_tag_list_from_bazefield(reg_ex_strings: list, key_vault_name: str):
    tag_list = get_full_tag_list_from_bazefield(key_vault_name)

    tag_ids = []

    for reg_ex in reg_ex_strings:
        tag_ids = tag_ids + match_taglist_to_regular_expression(tag_list=tag_list,
                                                                regular_expression=reg_ex)

    tag_list_to_download = {"tagIds": str(tag_ids)[1:-1].replace(" ", "")}

    tag_list_to_download["dateTimeFormat"] = "dd-MM-yyyy HH:mm:ss.fff"
    tag_list_to_download["calenderUnit"] = "Minute"
    tag_list_to_download["useAssetTitle"] = False
    tag_list_to_download["useInterval"] = True
    tag_list_to_download["exportInUtc"] = True
    return json.dumps(tag_list_to_download)

def prepare_download(from_timestamp: int,
                     key_vault_name: str,
                     aggregates: str,
                     interval: str,
                     reg_ex_strings: list,
                     download_file_resolution=1000*3600*24):

    from_timestamp_as_string = get_datetime_as_string(from_timestamp)
    end_timestamp = from_timestamp + download_file_resolution

    print("Downloading " + from_timestamp_as_string)

    url = at.get_secret_from_key_vault(key_vault_name, secret_name='bazefield-data-export-url')
    auth = get_bazefield_authentication_from_key_vault(key_vault_name)
    headers = get_bazefield_headers()
    proxy = {'https': at.get_secret_from_key_vault(key_vault_name, secret_name='equinor-proxy')}
    data = get_transformed_tag_list_from_bazefield(reg_ex_strings, key_vault_name)

    res = requests.post(url=url.format(from_timestamp, end_timestamp, interval, aggregates),
                        auth=auth,
                        headers=headers,
                        verify=False,
                        proxies=proxy,
                        data=data)

    filename = res.text
    next_timestamp = end_timestamp

    return filename, next_timestamp, from_timestamp_as_string

def download_file(filename: str, from_timestamp_as_string: str, key_vault_name: str):
    csv_name = from_timestamp_as_string + '.csv'
    filename = filename.replace(".txt", "")

    url = at.get_secret_from_key_vault(key_vault_name, secret_name='bazefield-get-file-url')
    auth = get_bazefield_authentication_from_key_vault(key_vault_name)
    proxy = {'https': at.get_secret_from_key_vault(key_vault_name, secret_name='equinor-proxy')}

    res = requests.get(url=url.format(filename),
                       auth=auth,
                       verify=False,
                       proxies=proxy)

    file = res.content.decode()

    destination = '../Data/UTC/'
    file_path = destination + csv_name

    with open(file_path, 'w') as f:
        f.write(file)
        print("Finished.")
        print("Downloaded to "+ str(file_path))

    df = pd.read_csv(file_path, sep=';')
    df = reduce_dataframe_size(df)

    df.to_csv(file_path, sep=';')

    return True


def download_data_from_bazefield_as_csv(from_timestamp: dt.datetime,
                                        to_timestamp: dt.datetime,
                                        aggregates: str,
                                        interval: str,
                                        reg_ex_strings: list,
                                        key_vault_name: str):
    """Downloads data from Bazefield using the Bazefield API.
    
    Arguments:
        from_timestamp {dt.datetime} -- Download data from and including this time.
        to_timestamp {dt.datetime} -- Download data to and including this time.
        aggregates {str} -- Which aggregates to download as specified by Bazefield.
        interval {str} -- Data interval to download. Example: '10Min'.
        reg_ex_strings {list} -- A list of regular expressions to match tags against.
        key_vault_name {str} -- KeyVault populated with secrets as described in repo.
    """
    from_timestamp_int = int(from_timestamp.timestamp()) * 1000
    to_timestamp_int = int(to_timestamp.timestamp())

    while True:
        try:
            filename, next_timestamp, from_timestamp_as_string = \
                prepare_download(from_timestamp=from_timestamp_int,
                                 key_vault_name=key_vault_name,
                                 aggregates=aggregates,
                                 interval=interval,
                                 reg_ex_strings=reg_ex_strings)
            print('Downloading ' + from_timestamp_as_string)
            print(filename)
            _ = download_file(filename, from_timestamp_as_string, key_vault_name)
        except Exception as e:
            print(e)
            from_timestamp_int = next_timestamp
            continue
        
        if dt.datetime.fromtimestamp(next_timestamp / 1000) \
           > dt.datetime.fromtimestamp(to_timestamp_int):
            return
        else:
            from_timestamp_int = next_timestamp



# pylint: skip-file
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aestene-lib", # Replace with your own username
    version="0.0.15",
    author="Arnt Erik Stene",
    author_email="steneae@gmail.com",
    description="Package containing reusable tools for different applications.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aestene/AesLib",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy==1.17.*',
        'seaborn==0.9.*',
        'matplotlib==3.1.*',
        'pandas==0.25.*',
        'azure-keyvault-secrets==4.0.*',
        'azure.identity',
        'azure-cli-core',
        'azure.mgmt.compute',
        'missingno==0.4.*']
)



from typing import List
import datetime as dt
import random

import pandas as pd
import numpy as np



import matplotlib.pyplot as plt

class IntervalAnalysis:
    def __init__(self, df_interval: pd.DataFrame):
        self.df_interval = df_interval

    # Merge overlapping intervals
    # intervals: List containing intervals sorted by start time
    # closed: 'closed' interval value for evt. new intervals
    @staticmethod
    def merge_overlapping_intervals(intervals: List, closed: str = 'neither')->List:
        assert len(intervals) > 0, 'Not enough data in "intervals"'
        merged = [intervals[0]]
        for interval in intervals:
            top = merged[-1]
            if top.overlaps(interval) and (interval.right > top.right):
                top = pd.Interval(left=top.left, right=interval.right, closed=closed)
            elif not top.overlaps(interval):
                merged.append(interval)
        return merged

    @classmethod
    def get_empty_intervals(cls, df: pd.DataFrame, cols: List = None)-> pd.DataFrame:
        freq = Utility.get_freq(df)

        res = pd.Series()
        cols = df.columns.to_list() if cols is None else cols

        for icol, colname in enumerate(cols):
            idx = df[colname].dropna().index
            
            # Compensate for leading nan
            if pd.isna(df.iloc[0, icol]):
                idx = idx.insert(0, df.index[0]-freq)
                
            ser1 = pd.Series(idx.tolist(), index=idx)
            ser2 = ser1.shift(-1)-ser1
            ser2 = ser2[ser2 != freq]
            res = res.append(ser2)
            
        res = res.dropna()

        # Group res by start time. Max length to be used to generate intervals.
        resgrp = res.groupby(level=0).agg(ucnt='nunique', valmax='max', valmin='min', cnt='count')
        # Generate intervals by (start_time, start_time + max(length))
        ser_interval = \
            resgrp.apply(lambda x: pd.Interval(x.name,
                                               x.name+x['valmax'],
                                               closed='neither'),
                         axis=1)
        merged = IntervalAnalysis.merge_overlapping_intervals(ser_interval.sort_values())

        # Create data frame with merged intervals.
        df_interval = pd.DataFrame(data={'interval': merged}, index=[x.left for x in merged])
        df_interval['duration'] = \
            df_interval.apply(lambda x: x['interval'].right - x['interval'].left, axis=1)
        df_interval['duration_min'] = \
            df_interval.apply(lambda x: x['duration'].total_seconds()/60, axis=1)

        return df_interval
    
    @staticmethod
    def plot_distribution(df_interval: pd.DataFrame,
                          cut_durations: dt.timedelta = None,
                          figsize: tuple = (16, 16),
                          bins=50):
        df = df_interval if cut_durations is None \
             else df_interval[df_interval.duration < cut_durations]
        plt.figure(figsize=figsize)
        df.duration_min.plot.hist(bins=bins)
        plt.gca().set(title="Distribution of missing data duration (min)",
                      xlabel="Duration, min",
                      ylabel="Frequency")
        plt.show()
    
    @staticmethod
    def get_timeline(df_interval: pd.DataFrame, df: pd.DataFrame) -> pd.DataFrame:
        df_timeline = pd.DataFrame(
            data={'dur':0},
            index=pd.DatetimeIndex(pd.date_range(
                start=df.index[0],
                end=df.index[-1],
                freq=Utility.get_freq(df)))).merge(
                    right=df_interval,
                    how='outer',
                    left_index=True,
                    right_index=True)
        df_timeline.loc[~pd.isna(df_timeline.duration_min), 'dur'] = \
            df_timeline.loc[~pd.isna(df_timeline.duration_min), 'duration_min']
        df_timeline.index = df_timeline.index.tz_localize(None)
        return df_timeline

    @staticmethod
    def plot_timeline(df_interval: pd.DataFrame, df: pd.DataFrame, figsize: tuple = (16, 16)):
        plt.figure(figsize=figsize)
        df_timeline = IntervalAnalysis.get_timeline(df_interval, df)
        df_timeline.dur.plot()
        plt.gca().set(title="Missing data duration (min) timeline",
                      xlabel="Time",
                      ylabel="Duration, min")
        plt.show()

        # df - original data set
        # df_interval: data frame containing 'empty' intervals, as returned by get_empty_intervals
        # method - Pandas interpolation method, e.g.:
                # method = 'linear'
                # method = 'pad'
                # method = 'time'
                # method = 'quadratic'
                # method = 'cubic'
        # missing_lengths - number of missing data points for each missing interval. E.g.:
                # missing_lengths = [x if x!=0  else 1 for x in range(0,21, 5) ]
        # nmissing = 15 #Number of missing intervals in each 'good' interval
        # min_neighbours = 1 #Guaranteed number of neighbours around missing


    @staticmethod
    def threshold_experiment(df: pd.DataFrame,
                             df_interval: pd.DataFrame,
                             method: str,
                             missing_lengths: List,
                             cols: List = None,
                             big_enough: dt.timedelta = dt.timedelta(days=10),
                             nmissing: int = 15,
                             min_neighbours: int = 1) -> pd.DataFrame:
        # find "Good" intervals
        df_good_interval = pd.DataFrame(
            data={'left': df_interval.interval.apply(lambda x: x.right),
                  'right':df_interval.shift(-1).interval.apply(lambda x: x.left
                                                               if not pd.isna(x)
                                                               else df.iloc[-1].name)})
        df_good_interval['duration'] = df_good_interval.right-df_good_interval.left
        # Iterate through 'good intervals' which are big enough
        # build a test dataframe with orig and interpolated columns
        # drop random data points

        periods = range(len(df_good_interval.loc[df_good_interval.duration >= big_enough]))
        cols_analyzed = df.columns.to_list() if cols is None else cols
        df_errs = pd.DataFrame(columns=cols_analyzed,
                               index=pd.MultiIndex.from_product([missing_lengths, periods],
                                                                names=['gap_length', 'period']))

        # Loop through columns
        for colname in cols_analyzed:

            for missing_length in missing_lengths:
                period = 0
                for row in df_good_interval.loc[df_good_interval.duration >= big_enough].itertuples():
                    df_test = pd.DataFrame(
                        data={'x': np.arange(len(df.loc[row.left:row.right, colname])),
                              'orig':df.loc[row.left:row.right, colname],
                              'interpolated': df.loc[row.left:row.right, colname]},
                        index=df.loc[row.left:row.right].index)
                    ndrops = \
                        sorted([random.randint(min_neighbours, \
                            len(df_test)-1-min_neighbours-missing_length)\
                            for x in range(nmissing)])
                    
                    # Check for and remove overlapping missing intervals
                    ndrops_corr = [ndrops[0]]
                    for ndrop in ndrops:
                        acur = ndrops_corr[-1]
                        if acur+missing_length > ndrop:
                            continue
                        else:
                            ndrops_corr.append(ndrop)
                            
                    ndrops = sorted([x+j for j in range(missing_length) for x in ndrops_corr])
                    df_test.iloc[ndrops, df_test.columns.get_loc('interpolated')] = np.nan
                    df_fixed = df_test['interpolated'].interpolate(method=method, axis=0)
                    df_errs.loc[(missing_length, period), colname] = \
                        Utility.mean_absolute_percentage_error(df_test.orig, df_fixed)
                    period += 1
        df_errs = df_errs.astype('float')
        return df_errs

    @staticmethod
    def plot_threshold_experiment(df_errs: pd.DataFrame, method: str, figsize: tuple = (16, 16)):
        df_errs_mean = df_errs.groupby('gap_length').mean()
        #df_errs_sem = df_errs.groupby('gap_length').sem()
        plt.figure()
        df_errs_mean.plot(figsize=figsize)
        plt.xlabel("Missing gap size, data points")
        plt.ylabel("MAPE, %")
        plt.suptitle(f"{method} Interpolation error (MAPE) by missing gap size")
        plt.show()


class Utility:

    @staticmethod
    def get_freq(df: pd.DataFrame)->dt.timedelta:
        assert len(df > 2), "Not enough data in the data set"

        check_level = None

        #Compensate for MultiIndex - find Datetime level
        if isinstance(df.index, pd.MultiIndex):
            for level in df.index.levels:
                if isinstance(level, pd.DatetimeIndex):
                    check_level = level
                    break
        elif isinstance(df.index, pd.DatetimeIndex):
            check_level = df.index

        assert check_level is not None, "Cannot define frequency - no DatetimeIndex given"
        return check_level[1]-check_level[0]

    @staticmethod
    def mean_absolute_percentage_error(y_true, y_pred):
        y_true, y_pred = np.array(y_true), np.array(y_pred)
        return np.mean(np.abs((y_true - y_pred) / y_true)) * 100
        


"""Tools to manipulate Pandas DataFrames for Data Science tasks.
"""
import pandas as pd

def set_rows_to_display_pandas(rows: int):
    """Set the number of rows Pandas will display when showing a DataFrame.

    Arguments:
        rows {int} -- Number of rows.
    """
    pd.options.display.max_rows = rows

def print_data_frame(df: pd.DataFrame, max_rows=5, max_columns=7, width=None):
    with pd.option_context(
            'display.max_rows', max_rows,
            'display.max_columns', max_columns,
            'display.width', width):
        print(df)

def get_idxmax_integer_pos_of_series(series: pd.Series) -> int:
    """Get the index of the largest entry in a Pandas series as an integer.

    Arguments:
        series {pd.Series} -- Series

    Returns:
        int -- Integer index of maximum value
    """
    current_index = 0
    max_index = 0
    max_value = 0
    for value in series:
        if value > max_value:
            max_value = value
            max_index = current_index
        current_index += 1
    return max_index

def read_csv_files(csv_files: list) -> list:
    csv_data_frames = []
    for filename in csv_files:
        df = pd.read_csv(filename, sep=';')
        csv_data_frames.append(df)
    return csv_data_frames

def stack_csv_files(csv_files: list, write_to_file=False) -> pd.DataFrame:
    csv_data_frames = read_csv_files(csv_files)
    stacked_data_frame = pd.concat(csv_data_frames, axis=0, ignore_index=True)
    stacked_data_frame.drop(stacked_data_frame.columns[0], axis=1, inplace=True)
    if write_to_file:
        stacked_data_frame.to_csv('stacked-dataframe.csv', sep=';', index=False)

    return stacked_data_frame

def check_for_columns_with_no_data_or_variation(df: pd.DataFrame):
    description = df.describe().T
    print("N Total columns         : {}".format(len(description)))
    print("N Cols with no data     : {}".format(len(description.loc[description['count'] == 0])))
    print("N Cols with no variation: {}".format(len(description.loc[description['std'] == 0])))

def count_all_columns(df: pd.DataFrame) -> pd.Series:
    return df.count()

def get_interval_size_with_consecutive_null(df: pd.DataFrame, column):
    intervals = \
        df[column].isnull().astype(int).groupby(df[column].notnull().astype(int).cumsum()).sum()
    return intervals

def remove_rows_with_zero_value(series: pd.Series):
    return series[series != 0]

def count_null_values_in_columns(df: pd.DataFrame):
    return df.isnull().sum()

def percentage_of_values_that_is_null(df: pd.DataFrame):
    return (df.isnull().sum() / len(df))*100

def show_columns_with_zero_variation(df: pd.DataFrame):
    description = df.describe()
    return description.loc[:, description.loc['std'] == 0]



from azure.keyvault.secrets import SecretClient
from azure.common.client_factory import get_client_from_cli_profile
from azure.mgmt.compute import ComputeManagementClient

def get_secret_from_key_vault(key_vault_name: str, secret_name: str):
    key_vault_uri = "https://" + key_vault_name + ".vault.azure.net"
    credential = get_client_from_cli_profile(ComputeManagementClient).config.credentials
    client = SecretClient(key_vault_uri, credential)

    return client.get_secret(secret_name).value

def set_secret_in_key_vault(key_vault_name: str, secret_name: str, secret_value):
    key_vault_uri = "https://" + key_vault_name + ".vault.azure.net"
    credential = get_client_from_cli_profile(ComputeManagementClient).config.credentials
    client = SecretClient(key_vault_uri, credential)
    client.set_secret(secret_name, secret_value)
